@@ -37,6 +37,7 @@ from collections import deque
 
 import sentry_sdk
 from sentry_sdk._compat import PY33, PY311
+from sentry_sdk._lru_cache import LRUCache
 from sentry_sdk._types import TYPE_CHECKING
 from sentry_sdk.utils import (
     capture_internal_exception,
@@ -65,19 +66,6 @@ if TYPE_CHECKING:
 
     ThreadId = str
 
-    # The exact value of this id is not very meaningful. The purpose
-    # of this id is to give us a compact and unique identifier for a
-    # raw stack that can be used as a key to a dictionary so that it
-    # can be used during the sampled format generation.
-    RawStackId = Tuple[int, int]
-
-    RawFrame = Tuple[
-        str,  # abs_path
-        int,  # lineno
-    ]
-    RawStack = Tuple[RawFrame, ...]
-    RawSample = Sequence[Tuple[str, Tuple[RawStackId, RawStack, Deque[FrameType]]]]
-
     ProcessedSample = TypedDict(
         "ProcessedSample",
         {
@@ -120,6 +108,21 @@ if TYPE_CHECKING:
         {"profile_id": str},
     )
 
+    FrameId = Tuple[
+        str,  # abs_path
+        int,  # lineno
+    ]
+    FrameIds = Tuple[FrameId, ...]
+
+    # The exact value of this id is not very meaningful. The purpose
+    # of this id is to give us a compact and unique identifier for a
+    # raw stack that can be used as a key to a dictionary so that it
+    # can be used during the sampled format generation.
+    StackId = Tuple[int, int]
+
+    ExtractedStack = Tuple[StackId, FrameIds, List[ProcessedFrame]]
+    ExtractedSample = Sequence[Tuple[ThreadId, ExtractedStack]]
+
 
 try:
     from gevent import get_hub as get_gevent_hub  # type: ignore
@@ -244,12 +247,16 @@ def teardown_profiler():
 MAX_STACK_DEPTH = 128
 
 
+CWD = os.getcwd()
+
+
 def extract_stack(
-    frame,  # type: Optional[FrameType]
-    prev_cache=None,  # type: Optional[Tuple[RawStackId, RawStack, Deque[FrameType]]]
+    raw_frame,  # type: Optional[FrameType]
+    cache,  # type: LRUCache
+    cwd=CWD,  # type: str
     max_stack_depth=MAX_STACK_DEPTH,  # type: int
 ):
-    # type: (...) -> Tuple[RawStackId, RawStack, Deque[FrameType]]
+    # type: (...) -> ExtractedStack
     """
     Extracts the stack starting the specified frame. The extracted stack
     assumes the specified frame is the top of the stack, and works back
@@ -259,38 +266,21 @@ def extract_stack(
     only the first `MAX_STACK_DEPTH` frames will be returned.
     """
 
-    frames = deque(maxlen=max_stack_depth)  # type: Deque[FrameType]
+    raw_frames = deque(maxlen=max_stack_depth)  # type: Deque[FrameType]
 
-    while frame is not None:
-        try:
-            f_back = frame.f_back
-        except AttributeError:
-            capture_internal_exception(sys.exc_info())
-            # For some reason, the frame we got isn't a `FrameType` and doesn't
-            # have a `f_back`. When this happens, we continue with any frames
-            # that we've managed to extract up to this point.
-            break
+    while raw_frame is not None:
+        f_back = raw_frame.f_back
+        raw_frames.append(raw_frame)
+        raw_frame = f_back
+
+    frame_ids = tuple(frame_id(raw_frame) for raw_frame in raw_frames)
+    frames = []
+    for i, fid in enumerate(frame_ids):
+        frame = cache.get(fid)
+        if frame is None:
+            frame = extract_frame(raw_frames[i], cwd)
+            cache.set(fid, frame)
         frames.append(frame)
-        frame = f_back
-
-    if prev_cache is None:
-        stack = tuple(frame_key(frame) for frame in frames)
-    else:
-        _, prev_stack, prev_frames = prev_cache
-        prev_depth = len(prev_frames)
-        depth = len(frames)
-
-        # We want to match the frame found in this sample to the frames found in the
-        # previous sample. If they are the same (using the `is` operator), we can
-        # skip the expensive work of extracting the frame information and reuse what
-        # we extracted during the last sample.
-        #
-        # Make sure to keep in mind that the stack is ordered from the inner most
-        # from to the outer most frame so be careful with the indexing.
-        stack = tuple(
-            prev_stack[i] if i >= 0 and frame is prev_frames[i] else frame_key(frame)
-            for i, frame in zip(range(prev_depth - depth, prev_depth), frames)
-        )
 
     # Instead of mapping the stack into frame ids and hashing
     # that as a tuple, we can directly hash the stack.
@@ -303,14 +293,14 @@ def extract_stack(
     # To Reduce the likelihood of hash collisions, we include
     # the stack depth. This means that only stacks of the same
     # depth can suffer from hash collisions.
-    stack_id = len(stack), hash(stack)
+    stack_id = len(raw_frames), hash(frame_ids)
 
-    return stack_id, stack, frames
+    return stack_id, frame_ids, frames
 
 
-def frame_key(frame):
-    # type: (FrameType) -> RawFrame
-    return (frame.f_code.co_filename, frame.f_lineno)
+def frame_id(raw_frame):
+    # type: (FrameType) -> FrameId
+    return (raw_frame.f_code.co_filename, raw_frame.f_lineno)
 
 
 def extract_frame(frame, cwd):
@@ -346,7 +336,7 @@ if PY311:
 
     def get_frame_name(frame):
         # type: (FrameType) -> str
-        return frame.f_code.co_qualname  # type: ignore
+        return frame.f_code.co_qualname
 
 else:
 
@@ -479,8 +469,8 @@ class Profile(object):
         self.stop_ns = 0  # type: int
         self.active = False  # type: bool
 
-        self.indexed_frames = {}  # type: Dict[RawFrame, int]
-        self.indexed_stacks = {}  # type: Dict[RawStackId, int]
+        self.indexed_frames = {}  # type: Dict[FrameId, int]
+        self.indexed_stacks = {}  # type: Dict[StackId, int]
         self.frames = []  # type: List[ProcessedFrame]
         self.stacks = []  # type: List[ProcessedStack]
         self.samples = []  # type: List[ProcessedSample]
@@ -620,8 +610,8 @@ class Profile(object):
 
         scope.profile = old_profile
 
-    def write(self, cwd, ts, sample, frame_cache):
-        # type: (str, int, RawSample, Dict[RawFrame, ProcessedFrame]) -> None
+    def write(self, ts, sample):
+        # type: (int, ExtractedSample) -> None
         if not self.active:
             return
 
@@ -637,31 +627,32 @@ class Profile(object):
 
         elapsed_since_start_ns = str(offset)
 
-        for tid, (stack_id, raw_stack, frames) in sample:
-            # Check if the stack is indexed first, this lets us skip
-            # indexing frames if it's not necessary
-            if stack_id not in self.indexed_stacks:
-                for i, raw_frame in enumerate(raw_stack):
-                    if raw_frame not in self.indexed_frames:
-                        self.indexed_frames[raw_frame] = len(self.indexed_frames)
-                        processed_frame = frame_cache.get(raw_frame)
-                        if processed_frame is None:
-                            processed_frame = extract_frame(frames[i], cwd)
-                            frame_cache[raw_frame] = processed_frame
-                        self.frames.append(processed_frame)
+        for tid, (stack_id, frame_ids, frames) in sample:
+            try:
+                # Check if the stack is indexed first, this lets us skip
+                # indexing frames if it's not necessary
+                if stack_id not in self.indexed_stacks:
+                    for i, frame_id in enumerate(frame_ids):
+                        if frame_id not in self.indexed_frames:
+                            self.indexed_frames[frame_id] = len(self.indexed_frames)
+                            self.frames.append(frames[i])
 
-                self.indexed_stacks[stack_id] = len(self.indexed_stacks)
-                self.stacks.append(
-                    [self.indexed_frames[raw_frame] for raw_frame in raw_stack]
+                    self.indexed_stacks[stack_id] = len(self.indexed_stacks)
+                    self.stacks.append(
+                        [self.indexed_frames[frame_id] for frame_id in frame_ids]
+                    )
+
+                self.samples.append(
+                    {
+                        "elapsed_since_start_ns": elapsed_since_start_ns,
+                        "thread_id": tid,
+                        "stack_id": self.indexed_stacks[stack_id],
+                    }
                 )
-
-            self.samples.append(
-                {
-                    "elapsed_since_start_ns": elapsed_since_start_ns,
-                    "thread_id": tid,
-                    "stack_id": self.indexed_stacks[stack_id],
-                }
-            )
+            except AttributeError:
+                # For some reason, the frame we get doesn't have certain attributes.
+                # When this happens, we abandon the current sample as it's bad.
+                capture_internal_exception(sys.exc_info())
 
     def process(self):
         # type: () -> ProcessedProfile
@@ -793,12 +784,7 @@ class Scheduler(object):
         # type: () -> Callable[..., None]
         cwd = os.getcwd()
 
-        # In Python3+, we can use the `nonlocal` keyword to rebind the value,
-        # but this is not possible in Python2. To get around this, we wrap
-        # the value in a list to allow updating this value each sample.
-        last_sample = [
-            {}
-        ]  # type: List[Dict[int, Tuple[RawStackId, RawStack, Deque[FrameType]]]]
+        cache = LRUCache(max_size=256)
 
         def _sample_stack(*args, **kwargs):
             # type: (*Any, **Any) -> None
@@ -810,7 +796,6 @@ class Scheduler(object):
             if not self.new_profiles and not self.active_profiles:
                 # make sure to clear the cache if we're not profiling so we dont
                 # keep a reference to the last stack of frames around
-                last_sample[0] = {}
                 return
 
             # This is the number of profiles we want to pop off.
@@ -825,16 +810,16 @@ class Scheduler(object):
 
             now = nanosecond_time()
 
-            raw_sample = {
-                tid: extract_stack(frame, last_sample[0].get(tid))
-                for tid, frame in sys._current_frames().items()
-            }
-
-            # make sure to update the last sample so the cache has
-            # the most recent stack for better cache hits
-            last_sample[0] = raw_sample
-
-            sample = [(str(tid), data) for tid, data in raw_sample.items()]
+            try:
+                sample = [
+                    (str(tid), extract_stack(frame, cache, cwd))
+                    for tid, frame in sys._current_frames().items()
+                ]
+            except AttributeError:
+                # For some reason, the frame we get doesn't have certain attributes.
+                # When this happens, we abandon the current sample as it's bad.
+                capture_internal_exception(sys.exc_info())
+                return
 
             # Move the new profiles into the active_profiles set.
             #
@@ -851,11 +836,9 @@ class Scheduler(object):
 
             inactive_profiles = []
 
-            frame_cache = {}  # type: Dict[RawFrame, ProcessedFrame]
-
             for profile in self.active_profiles:
                 if profile.active:
-                    profile.write(cwd, now, sample, frame_cache)
+                    profile.write(now, sample)
                 else:
                     # If a thread is marked inactive, we buffer it
                     # to `inactive_profiles` so it can be removed.
